package java.awt;
import java.awt.image.*;
import java.awt.event.*;
import java.io.*;
import java.awt.peer.*;
import java.beans.*;
import java.util.Locale;
import java.util.Vector;
import java.util.EventListener;

public abstract class Component
    implements ImageObserver, MenuContainer, Serializable {

    public static final float TOP_ALIGNMENT    = 0.0F;
    public static final float CENTER_ALIGNMENT = 0.5F;
    public static final float BOTTOM_ALIGNMENT = 1.0F;
    public static final float LEFT_ALIGNMENT   = 0.0F;
    public static final float RIGHT_ALIGNMENT  = 1.0F;

    /* FIXME: Only one global tree lock for the all component?
       Considering that there can exist completely seperate component
       trees on different graphics devices, it doesn't seem right that
       actions on one graphics device should defer layout on other
       graphics devives. */
    /* Make the treelock a String so that it can easily be identified
    in debug dumps. */
    private static Object AWT_TREE_LOCK = "AWT_TREE_LOCK";
    
    /* TODO: there is a lot of fields here...  should consider moving
       seldom used fields into an auxillary object that is only
       created when needed. */
    private String name;

    /** The java.awt.Container.addImpl() method manipulates this field
        directly. */
    Container parent;

    /* volatile since the event thread might be waiting for this to
       be set by the constructor so that it can start delivering paint
       events. */
    volatile ComponentPeer peer;

    private boolean isValid = false;
    private boolean isVisible = true; // default (except for Window)
    private boolean isEnabled = true;
    private Color foreground;
    private Color background;
    private Font font;
    private int x, y, width, height;
    private Locale locale;
    
    GraphicsConfiguration config;

    
    // FIXME: Implement AWTMulticaster instead
    Vector mouseListeners = new Vector();

    protected Component() {
    }

    public String getName() {
	return name;
    }
    
    public void setName(String name) {
	this.name = name;
    }

    public Container getParent() {
	return parent;
    }

    /** This method is synchronized so that we can prevent other
	threads from accessing the peer before addNotify() has
	completed. Implementations of addNotify() that sets the peer
	property should ensure that creation of the peer and setting
	of the peer property is executed in synchronized block. This
	will prevent events generated by the creation of the peer from
	causing peer property accesses before the property has been
	propertly set.
    */
    public synchronized ComponentPeer getPeer() {
	return peer;
    }

    /** This method must be called within the same synchronized block
        that creates the peer. */
    public synchronized void setPeer(ComponentPeer peer) {
	this.peer = peer;
    }

        


    //public void setDropTarget(DropTarget target);
    //public DropTarget getDropTarget();
    public GraphicsConfiguration getGraphicsConfiguration() {
	ComponentPeer peer = getPeer();
	if (peer != null) {
	    GraphicsConfiguration config = peer.getGraphicsConfiguration();
	    if (config != null) return config;
	}

	Container parent = getParent();
	if (parent != null) return parent.getGraphicsConfiguration();
	    
	return null;
    }
    

    public final Object getTreeLock() {
	return AWT_TREE_LOCK;
    }

    public Toolkit getToolkit() {
	ComponentPeer peer = getPeer();
	if (peer != null) {
	    Toolkit tk = peer.getToolkit();
	    if (tk != null) return tk;
	}
	Container parent = getParent();
	if (parent != null) return parent.getToolkit();
	return Toolkit.getDefaultToolkit();
    }

    public boolean isValid() {
	return isValid;
    }
    

    public boolean isDisplayable() {
	Container parent = getParent();
	if (parent != null) return parent.isDisplayable();
	return false;
    }

    public boolean isVisible() {
	return isVisible;
    }
    
    public boolean isShowing() {
	Container parent = getParent();
	if (parent == null) return false;
	return parent.isShowing() && isVisible();
    }
    
    public boolean isEnabled() {
	return isEnabled;
    }

    public void setEnabled(boolean enabled) {
	isEnabled = enabled;
    }    

    // Deprecated methods...
    //public void enable();
    //public void enable(boolean b)
    //public void disable()
    //public void show();
    //public void show(boolean visible);
    //public void hide();

    public boolean isDoubleBuffered() {
	return false;
    }
    
    public void enableInputMethods(boolean enable) {
	throw new UnsupportedOperationException("not implemented");
    }

    public void setVisible(boolean visible) {
	isVisible = visible;
	ComponentPeer peer = getPeer();
	if (peer != null) peer.setVisible(isVisible);
    }

    public Color getForeground() {
	if (foreground != null) return foreground;
	Container parent = getParent();
	if (parent != null) return parent.getForeground();
	
	// probably wrong...
	throw new IllegalComponentStateException("no foreground yet");	
    }
    
    public void setForeground(Color fg) {
	foreground = fg;
	ComponentPeer peer = getPeer();
	if (peer != null) peer.setForeground(foreground);
    }

    /** @return the background color of the component. null may be
        returned instead of the actual background color, if this
        method is called before the component is added to the
        component hierarchy. */
    public Color getBackground() {
	if (background != null) return background;
	Container parent = getParent();
	if (parent != null) return parent.getBackground();

	/* compatibility testing has shown that the correct thing to
	   do here is simply to return null. Unlike getLocale() this
	   method should not throw IllegalComponentStateException if
	   the correct value is not known yet. */
	return null;
    }
    
    /** Specifies the color that will be used to fill the background
        of a component before normal painting commences. */
    public void setBackground(Color bg) {
	background = bg;
	ComponentPeer peer = getPeer();
	if (peer != null) peer.setBackground(background);
    }
    
    public Font getFont() {
	if (font != null) return font;
	Container parent = getParent();
	if (parent != null) return parent.getFont();

	// probably wrong...
	throw new IllegalComponentStateException("no font yet");
    }

    public void setFont(Font font) {
	this.font = font;
	ComponentPeer peer = getPeer();
	peer.setFont(font);
    }
    
    public Locale getLocale() {
	if (locale != null) return locale;
	Container parent = getParent();
	if (parent != null) return parent.getLocale();

	throw new IllegalComponentStateException("no locale yet");
    }
    
    public void setLocale(Locale locale) {
	this.locale = locale;
	
	/* new writing/layout direction perhaps, or make more/less
	   room for localized text labels */
	invalidate();
    }

    /** This method should for most purposes be useless, since there
	really isn't much use in knowing the color model of a
	component. Slightly more interesting is
	getGraphicsConfiguration(). */
    public ColorModel getColorModel() {
	GraphicsConfiguration config = getGraphicsConfiguration();

	if (config != null) return config.getColorModel();

	return getToolkit().getColorModel();
    }

    public Point getLocation() {
	return new Point(x, y);
    }

    public Point getLocationOnScreen() {
	throw new UnsupportedOperationException("not implemented");
    }

    // deprecated;
    //public Point location();
    //public void move(int x, int y);
    //public Dimension size();
    //public void resize(int width, int height);
    //public void resize(Dimension d);
    //public Rectangle bounds();
    //public void reshape(int x, int y, int width, int height);

    public void setLocation(int x, int y) {
	setBounds(x, y, width, height);
    }

    public void setLocation(Point p) {
	setLocation(x, y);
    }

    public Dimension getSize() {
	return new Dimension(width, height);
    }
    
    public void setSize(int width, int height) {
	setBounds(x, y, width, height);
    }

    public void setSize(Dimension size) {
	setSize(size.width, size.height);
    }

    public Rectangle getBounds() {
	return new Rectangle(x, y, width, height);
    }

    public void setBounds(int x, int y, int width, int height) {

	if ((this.x == x) && (this.y == y) &&
	    (this.width == width) && (this.height == height)) return;
	this.x = x;
	this.y = y;
	this.width = width;
	this.height = height;

	ComponentPeer peer = getPeer();
	if (peer != null) peer.setBounds(x, y, width, height);
    }
    
    public void setBounds(Rectangle bounds) {
	setBounds(bounds.x, bounds.y, bounds.width, bounds.height);
    }
    
    public int getX() {
	return x;
    }

    public int getY() {
	return y;
    }
    
    public int getWidth() {
	return width;
    }

    public int getHeight() {
	return height;
    }
    
    public Rectangle getBounds(Rectangle bounds) {
	bounds.x = x;
	bounds.y = y;
	bounds.width = width;
	bounds.height = height;
	return bounds;
    }
    
    public Dimension getSize(Dimension size) {
	size.width = width;
	size.height = height;
	return size;
    }
    
    public Point getLocation(Point point) {
	point.x = x;
	point.y = y;
	return point;
    }

    public boolean isOpaque() {
	return !isLightweight();
    }

    /** Note that this also returns true if there currently is no peer
        object. */
    public boolean isLightweight() {
	return (getPeer() instanceof LightweightPeer);
    }
    
    public Dimension getPreferredSize() {
	ComponentPeer peer = getPeer();
	if (peer != null) {
	    Dimension size = peer.getPreferredSize();
	    if (size != null) return size;
	}
	
	return getSize();
    }

    //public Dimension preferredSize();

    public Dimension getMinimumSize() {
	ComponentPeer peer = getPeer();
	if (peer != null) return peer.getMinimumSize();
	return getPreferredSize();
    }

    // deprecated:
    //public Dimension minimumSize();
    //public void layout();

    public Dimension getMaximumSize() {
	return getPreferredSize();
    }
    

    public float getAlignmentX() {
	throw new UnsupportedOperationException("not implemented");
    }

    public float getAlignmentY() {
	throw new UnsupportedOperationException("not implemented");
    }

    /** Normally never called on components. Is called on containers
        during validation. */
    public void doLayout() {
	// nothing to unless we're a container
    }


    public void validate() {
	if (isValid()) return;

	validateTree();
    }
    
    void validateTree() {
	isValid = true;
    }


    /** Make sure to implement this effectively in subclasses. 
	Atleast do a check...
	<pre>
	if (!isValid()) return;
	</pre>
    */
    public void invalidate() {
	isValid = false;
    }


    public Graphics getGraphics() {
	ComponentPeer peer = getPeer();
	
	if (peer != null) {
	    Graphics gfx = peer.getGraphics();
	    if (gfx != null) return gfx;

	    // assume lightweight
	    Container parent = getParent();
	    if (parent != null) {
		gfx = parent.getGraphics();
		Rectangle bounds = getBounds();
		gfx.setClip(bounds);
		gfx.translate(bounds.x, bounds.y);
		return gfx;
	    }
	}
	throw new IllegalStateException("no peer");
    }

    //public FontMetrics getFontMetrics(Font font);
    //public void setCursor(Cursor cursor);
    //public Cursor getCursor();

    public void paint(Graphics gfx) {
    }

    public void update(Graphics gfx) {

	// background clearing 
	//gfx.clearRect(0, 0, width, height);
	paint(gfx);

    }

    public void paintAll(Graphics gfx) {
	if (!isVisible()) return;
	
	ComponentPeer peer = getPeer();
	if (peer != null) peer.paint(gfx);
	paint(gfx);
    }
    

    public void repaint() {
	repaint(0, 0, getWidth(), getHeight());
    }
    
    public void repaint(long time) {
	// temporary implementation
	repaint();
    }

    public void repaint(int x, int y, int width, int height) {

	// Handle lightweight repainting by forwarding to native parent
	if (isLightweight()) {
	    Container parent = getParent();
	    if (parent != null)
		parent.repaint(x+getX(), y+getY(), width, height);
	    return;
	}
		
	ComponentPeer peer = getPeer();
	if (peer != null) peer.repaint(x, y, width, height);
    }
    
    public void repaint(long tm, int x, int y, int width, int height) {
	// temporary implementation
	repaint(x, y, width, height);
    }

    public void print(Graphics gfx) {
	paint(gfx);
    }

    public void printAll(Graphics gfx) {
	paintAll(gfx);
    }

    public boolean imageUpdate(Image img, int infoflags,
			       int x, int y, int w, int h) {
	throw new UnsupportedOperationException("not implemented");
    }

    public Image createImage(ImageProducer producer) {
	throw new UnsupportedOperationException("not implemented");
    }

    public Image createImage(int width, int height) {
	return getGraphicsConfiguration().createCompatibleImage(width, height);
    }

    public boolean prepareImage(Image image, ImageObserver observer) {
	throw new UnsupportedOperationException("not implemented");
    }
	

    public boolean prepareImage(Image image, int width, int height,
				ImageObserver observer) {
	throw new UnsupportedOperationException("not implemented");
    }

    public int checkImage(Image image, ImageObserver observer) {
	throw new UnsupportedOperationException("not implemented");
    }

    public int checkImage(Image image, int width, int height,
			  ImageObserver observer) {
	throw new UnsupportedOperationException("not implemented");
    }


    /** Tests to see if component contains the specified point. The
        coordinates are given in the coordinate system of this
        component. */

    public boolean contains(int x, int y) {
	return (x>=0) && (y>=0) && (x<getWidth()) && (y<getHeight());
    }

    //public boolean inside(int x, int y);

    public boolean contains(Point p) {
	//return getBounds().contains(p);
	return true;
    }

    /** Returns the immediate child at the specified location, if
	any. Coordinates are given in the coordinate system of the
	component.

	@returns immediate child if found, null if coordinates lie
	outside component, or the component itself if no child was
	found, but coordinates is within component.  */
    public Component getComponentAt(int x, int y) {
	if (contains(x, y)) return this;
	return null;
    }

    //public Component locate(int x, int y);

    public Component getComponentAt(Point p) {
	if (contains(p)) return this;
	return null;
    }
    
    //public void deliverEvent(Event evt);

    public final void dispatchEvent(AWTEvent evt) {
	processEvent(evt);
	
    }    

    //public boolean postEvent(Event evt);

    /*
    public void addComponentListener(ComponentListener l) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    public void removeComponentListener(ComponentListener l) {		    
	throw new UnsupportedOperationException("not implemented");
    }
    
    public void addFocusListener(FocusListener l) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    public void removeFocusListener(FocusListener l) {
	throw new UnsupportedOperationException("not implemented");
    }

    public void addHierarchyListener(HierarchyListener l) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    public void removeHierarchyListener(HierarchyListener l) {
	throw new UnsupportedOperationException("not implemented");
    }
    

    public void addHierarchyBoundsListener(HierarchyBoundsListener l) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    public void removeHierarchyBoundsListener(HierarchyBoundsListener l) {
	throw new UnsupportedOperationException("not implemented");
    }
    */

    public void addKeyListener(KeyListener l) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    public void removeKeyListener(KeyListener l) {
	throw new UnsupportedOperationException("not implemented");
    }

    public void addMouseListener(MouseListener l) {
	mouseListeners.addElement(l);
	// FIXME: notify peer, in case the event mask may need to be changed
    }

    public void removeMouseListener(MouseListener l) {
	mouseListeners.removeElement(l);
	// FIXME: notify peer, in case the event mask may need to be changed
    }

    public void addMouseMotionListener(MouseMotionListener l) {
	throw new UnsupportedOperationException("not implemented");
    }

    public void removeMouseMotionListener(MouseMotionListener l) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    public void addInputMethodListener(InputMethodListener l) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    public void removeInputMethodListener(InputMethodListener l) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    public EventListener[] getListeners(Class listenerType) {
	if (listenerType == MouseListener.class) {
	    MouseListener[] arr = new MouseListener[mouseListeners.size()];
	    mouseListeners.copyInto(arr);
	    return arr;
	}
	
	return null;
    }
    

    //public InputMethodRequests getInputMethodRequests();

    //public InputContext getInputContext();

    protected final void enableEvents(long enableFlags) {
	throw new UnsupportedOperationException("not implemented");
    }

    protected final void disableEvents(long disableFlags) {
	throw new UnsupportedOperationException("not implemented");
    }

    protected AWTEvent coalesceEvents(AWTEvent queuedEvent,
				      AWTEvent newEvent) {
	switch (newEvent.getID()) {
	case MouseEvent.MOUSE_MOVED:
	case MouseEvent.MOUSE_DRAGGED:
	    return newEvent;
	case PaintEvent.PAINT:
	case PaintEvent.UPDATE:

	    return coalescePaintEvents((PaintEvent) queuedEvent,
				       (PaintEvent) newEvent);


	    // FIXME: should keep a refresh-list
	}
	
	return null; // couldn't combine events
    }
    
    private PaintEvent coalescePaintEvents(PaintEvent queuedEvent,
					   PaintEvent newEvent) {
	Rectangle r1 = queuedEvent.getUpdateRect();
	Rectangle r2 = newEvent.getUpdateRect();
	Rectangle union = r1.union(r2);

	int r1a = r1.width * r1.height;
	int r2a = r2.width * r2.height;
	int ua  = union.width * union.height;

	if (ua > (r1a+r2a)*2) return null;

	newEvent.setUpdateRect(union);
	return newEvent;
    }


    protected void processEvent(AWTEvent evt) {
	
	if (evt instanceof MouseEvent) {
	    processMouseEvent((MouseEvent) evt);
	}


	ComponentPeer peer = getPeer();
	if (peer != null) {
	    
	    /* Let peer take care of things such as painting, etc. */

	    peer.processEvent(evt);
	}
	

	if (evt instanceof PaintEvent) {

	    // Can't do graphics without peer
	    if (peer == null) {
		System.err.println("detached paint event ignored ("+this+")");
		return;
	    }

	    PaintEvent pEvt = (PaintEvent) evt;
	    Graphics gfx = getGraphics();
	    Shape clip = pEvt.getUpdateRect();
	    gfx.setClip(clip);

	    switch (evt.getID()) {
	    case PaintEvent.PAINT:
		if (peer != null) paint(gfx);
		break;
	    case PaintEvent.UPDATE:
		if (peer != null) update(gfx);
		break;
	    default:
		throw new RuntimeException("unknown paint event");
	    }
	} 
    }
    
    protected void processComponentEvent(ComponentEvent evt){
	throw new UnsupportedOperationException("not implemented");
    }

    protected void processFocusEvent(FocusEvent evt) {
	throw new UnsupportedOperationException("not implemented");
    }

    protected void processKeyEvent(KeyEvent evt) {
	throw new UnsupportedOperationException("not implemented");
    }
    

    /** Process a mouse event by delivering it to all interested
        listeners. This method assumes that the event already has been
        redirected to the appropriate lightweight component. */
    protected void processMouseEvent(MouseEvent event) {

	// OK, we've been hit...

	MouseListener[] listeners = (MouseListener[])
	    getListeners(MouseListener.class);
	for (int i=0; i<listeners.length; i++) {
	    switch (event.getID()) {
	    case MouseEvent.MOUSE_PRESSED:
		listeners[i].mousePressed(event);
		break;	
	    case MouseEvent.MOUSE_RELEASED:
		listeners[i].mouseReleased(event);
		break;
	    case MouseEvent.MOUSE_CLICKED:
		listeners[i].mouseClicked(event);
		break;
	    case MouseEvent.MOUSE_ENTERED:
		listeners[i].mouseEntered(event);
		break;
	    case MouseEvent.MOUSE_EXITED:
		listeners[i].mouseExited(event);
		break;
	    default:
		throw new Error("event not intended for MouseListeners");
	    }
	}
    }
    
    protected void processMouseMotionEvent(MouseEvent evt) {
	throw new UnsupportedOperationException("not implemented");
    }

    protected void processInputMethodEvent(InputMethodEvent evt) {
	throw new UnsupportedOperationException("not implemented");
    }

    /*
    protected void processHierarchyEvent(HierarchyEvent evt) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    protected void processHierarchyBoundsEvent(HierarchyEvent evt) {
	throw new UnsupportedOperationException("not implemented");
    }
    */

    // Deprecated:
    //public boolean handleEvent(Event evt);
    //public boolean mouseDown(Event evt, int x, int y);
    //public boolean mouseDrag(Event evt, int x, int y);
    //public boolean mouseUp(Event evt, int x, int y);
    //public boolean mouseMove(Event evt, int x, int y);
    //public boolean mouseEnter(Event evt, int x, int y);
    //public boolean mouseExit(Event evt, int x, int y);
    //public boolean keyDown(Event evt, int key);
    //public boolean keyUp(Event evt, int key);
    //public boolean action(Event evt, Object what);

    public void addNotify() {
	if (peer == null) {
	    synchronized (this) {
		setPeer(getToolkit().createComponent(this));
	    }
	}

	/* We do not invalidate here, but rather leave that job up to
	   the peer. For efficiency, the peer can choose not to
	   invalidate if it is happy with the current dimensions,
	   etc. */

    }

    public void removeNotify() {
	peer = null;
    }

    //public boolean gotFocus(Event evt, Object what);
    //public boolean lostFocus(Event evt, Object what);

    public boolean isFocusTraversable() {
	throw new UnsupportedOperationException("not implemented");
    }

    public void requestFocus() {
	throw new UnsupportedOperationException("not implemented");
    }

    public void transferFocus() {
	throw new UnsupportedOperationException("not implemented");
    }

    //public void nextFocus()

    public boolean hasFocus() {
	throw new UnsupportedOperationException("not implemented");
    }

    /*
    public void add(PopupMenu popup) {
	throw new UnsupportedOperationException("not implemented");
    }

    public void remove(MenuComponent popup) {
	throw new UnsupportedOperationException("not implemented");
    } */

    protected String paramString() {
	StringBuffer param = new StringBuffer();
	String name = getName();
	if (name != null) {
	    param.append(name);
	    param.append(",");
	}
	Rectangle bounds = getBounds();
	param.append(bounds.width);
	param.append("x");
	param.append(bounds.height);
	param.append("+");
	param.append(bounds.x);
	param.append("+");
	param.append(bounds.y);
	
	if (!isValid()) param.append(",invalid");
	if (!isVisible()) param.append(",invisible");
	if (!isEnabled()) param.append(",disabled");
	if (!isOpaque()) param.append(",translucent");
	if (isDoubleBuffered()) param.append(",doublebuffered");
	
	return param.toString();
    }

    public String toString() {
	return getClass().getName() + "[" + paramString() + "]";
    }

    public void list() {
	list(System.out);
    }

    public void list(PrintStream out) {
	list(out, 0);
    }

    public void list(PrintStream out, int indent) {
	throw new UnsupportedOperationException("not implemented");
    }

    public void list(PrintWriter out) {
	list(out, 0);
    }

    public void list(PrintWriter out, int indent) {
	throw new UnsupportedOperationException("not implemented");
    }

    public void addPropertyChangeListener(PropertyChangeListener listener) {
	throw new UnsupportedOperationException("not implemented");
    }

    public void removePropertyChangeListener(PropertyChangeListener listener) {
	throw new UnsupportedOperationException("not implemented");
    }

    public void addPropertyChangeListener(String propertyName,
					  PropertyChangeListener listener) {
	throw new UnsupportedOperationException("not implemented");
    }

    public void removePropertyChangeListener(String propertyName,
					     PropertyChangeListener listener) {
	throw new UnsupportedOperationException("not implemented");
    }

    protected void firePropertyChange(String propertyName,
				      Object oldValue,
				      Object newValue) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    //public void setComponentOrientation(ComponentOrientation o)
    //public ComponentOrientation getComponentOrientation()
    //public AccessibleContext getAccessibleContext()


    /**
     *  This method is package private and is not a part of the AWT
     *  API. To access it outside the java.awt package do:
     *  <pre>
     *  no.rwr.PeerHints hints = (no.rwr.PeerHints)
     *      Toolkit.getProperty("no.rwr.PeerHints");
     *  hints.getEnabledEvents(component);
     *  </pre>
     */
    //long getEnabledEvents... aw, forget it...

}
